unit uProcess;

interface

uses ComCtrls, SysUtils;

procedure Encrypt(const SourceFile, SourceBitmap, Destination: string; BitsPerChannel: LongInt; ProgressBar: TProgressBar);
procedure Decrypt(const SourceFile, DestFile: string; ProgressBar: TProgressBar);

implementation

uses Windows, Graphics, Classes, uBits, Forms;

// One byte for every channel of the RGB trio
type pRGBArray= ^TRGBArray;
     TRGBArray= array [1..3] of Byte;

procedure ProcessEncrypt(Bitmap: TBitmap; Source: TFileStream; Destination: string; BPC: LongInt; ProgressBar: TProgressBar);
var SourceIndex, SourceSize: LongInt;
    BitIndex, PixelBitIndex: LongInt;
    SourceByte: Byte;
    PixelsRow: pRGBArray;
    RGBIndex: Integer;
    PixelsRowMax, PixelsRowIndex, CurrentRow: Integer;

// A procedure inside another is quite ugly, but we avoid passing a lot of parameters when we call it
procedure CheckNextPixel;
begin
  if (RGBIndex <= 3) and (PixelBitIndex + 1 < BPC) then // We're OK, go for the next bit
    Inc(PixelBitIndex)
  else if RGBIndex < 3 then // Switch to next RGB channel
    begin
    Inc(RGBIndex);
    PixelBitIndex:= 0;
    end
  else if RGBIndex = 3 then // Load next pixel
    begin
    PixelBitIndex:= 0;
    RGBIndex:= 1;
    if PixelsRowIndex = PixelsRowMax then // We used all pixels in this row
      begin
      Inc(CurrentRow);
      PixelsRow:= Bitmap.ScanLine[CurrentRow];
      PixelsRowIndex:= 1;
      end
    else  // We still have pixels left in this row
      begin
      Inc(PixelsRowIndex);
      Inc(PixelsRow); // Increment the pointer so it points to the next pixel
      end;
    end;
end;

begin {ProcessEncrypt}

  {-- STORE THE BITS PER CHANNEL VALUE--}

  PixelsRow:= Bitmap.ScanLine[0];

  // We use 2 bits in the B channel to bring the count to 4 bits, so we can store BPC = 8, value that usually brings
  // devastation to the image quality...
  SetBitAt(PixelsRow^[1], 0, GetBitAt(BPC, 0));
  SetBitAt(PixelsRow^[1], 1, GetBitAt(BPC, 1));
  SetBitAt(PixelsRow^[2], 0, GetBitAt(BPC, 2));
  SetBitAt(PixelsRow^[3], 0, GetBitAt(BPC, 3));

  // Initialize
  PixelsRowMax:= Bitmap.Width;
  CurrentRow:= 0;
  PixelBitIndex:= 0;
  PixelsRowIndex:= 2;
  RGBIndex:= 1;
  Inc(PixelsRow);

  {-- STORE THE ACTUAL LENGTH OF THE DATA --}

  SourceSize:= Source.Size;
  for BitIndex:= 0 to SizeOf(SourceSize) * 8 - 1 do
    begin
    SetBitAt(PixelsRow^[RGBIndex], PixelBitIndex, GetBitAt(SourceSize, BitIndex));
    CheckNextPixel;
    end;

  {-- STORE THE DATA --}

  // Copy the bits from every byte in the source stream onto the bits in the pixels
  Source.Seek(0, soFromBeginning);
  for SourceIndex:= 0 to SourceSize - 1 do
    begin

    if (SourceIndex + 1) mod 10 = 0 then
      begin
      ProgressBar.StepIt;
      Application.ProcessMessages;
      end;

    Source.Read(SourceByte, 1);
    for BitIndex:= 0 to 7 do
      begin
      SetBitAt(PixelsRow^[RGBIndex], PixelBitIndex, GetBitAt(SourceByte, BitIndex));
      CheckNextPixel;
      end;

    end;// for SourceIndex
end;

procedure Encrypt(const SourceFile, SourceBitmap, Destination: string; BitsPerChannel: LongInt; ProgressBar: TProgressBar);
var Bitmap: TBitmap;
    sSource: TFileStream;

begin
  ProgressBar.Position:= 0;
  ProgressBar.Step:= 1;

  Bitmap:= TBitmap.Create;
  sSource:= nil;
  try
    Bitmap.LoadFromFile(SourceBitmap);

    if Bitmap.PixelFormat <> pf24bit then
      raise Exception.Create('The image must have a 24-bit depth.');

    sSource:= TFileStream.Create(SourceFile, fmOpenRead);

    if sSource.Size = 0 then
      raise Exception.Create('The source file is 0 bytes. There''s nothing to hide.');

    if sSource.Size * 8 + SizeOf(LongInt) * 8  + 1 > Bitmap.Width * Bitmap.Height * 3 * BitsPerChannel then
      raise Exception.Create('The image is not big enough to accommodate the file.');
    // + 1: The BitsPerChannel value is stored in the first pixel
    // SizeOf(LongInt) * 8: we'll store the amount of actual encrypted data in the first pixels after BitsPerChannel

    ProgressBar.Max:= sSource.Size div 10;

    ProcessEncrypt(Bitmap, sSource, Destination, BitsPerChannel, ProgressBar);

    Bitmap.SaveToFile(Destination);

  finally
    Bitmap.Free;
    if Assigned(sSource) then sSource.Free;
  end;

end;

{----------------------------------------------------------------------------------------------------------------------}

procedure ProcessDecrypt(Bitmap: TBitmap; Destination: TFileStream; ProgressBar: TProgressBar);
var DataSize, DataIndex: LongInt;
    Data, BitIndex: Byte;
    PixelsRow: pRGBArray;
    PixelsRowMax, PixelsRowIndex, CurrentRow, MaxRows: Integer;
    PixelBitIndex: LongInt;
    RGBIndex: Integer;
    BPC: LongInt;

// A procedure inside another (and basically the same as the one in ProcessEncrypt) is quite ugly, but we avoid passing
// a lot of parameters when we call it
procedure CheckNextPixel;
begin
  if (RGBIndex <= 3) and (PixelBitIndex  + 1 < BPC) then // We're OK, go for the next bit
    Inc(PixelBitIndex)
  else if RGBIndex < 3 then // Switch to next RGB channel
    begin
    Inc(RGBIndex);
    PixelBitIndex:= 0;
    end
  else if RGBIndex = 3 then // Load next pixel
    begin
    PixelBitIndex:= 0;
    RGBIndex:= 1;
    if PixelsRowIndex = PixelsRowMax then // We used all pixels in this row
      begin
      Inc(CurrentRow);
      if CurrentRow > MaxRows then
        raise Exception.Create('The end of the image was reached while trying to read the hidden information.' + #13#10+
                               'This is probably caused by an image that doesn''t contain any hidden data.');

      PixelsRow:= Bitmap.ScanLine[CurrentRow];
      PixelsRowIndex:= 1;
      end
    else  // We still have pixels left in this row
      begin
      Inc(PixelsRowIndex);
      Inc(PixelsRow); // Increment the pointer so it points to the next pixel
      end;
    end;
end;

begin {ProcessDecrypt}

  {-- GET THE BITS PER CHANNEL VALUE --}

  PixelsRow:= Bitmap.ScanLine[0];
  BPC:= 0;
  SetBitAt(BPC, 0, GetBitAt(PixelsRow^[1], 0));
  SetBitAt(BPC, 1, GetBitAt(PixelsRow^[1], 1));
  SetBitAt(BPC, 2, GetBitAt(PixelsRow^[2], 0));
  SetBitAt(BPC, 3, GetBitAt(PixelsRow^[3], 0));

  if (BPC < 1 ) or (BPC > 8) then
    raise Exception.Create('The BitsPerChannel value is not in the range 1-8.' + #13#10 +
                           'This is probably caused by an image that doesn''t contain any hidden data.');

  // Initialize
  PixelsRowMax:= Bitmap.Width;
  MaxRows:= Bitmap.Height - 1;
  CurrentRow:= 0;
  PixelBitIndex:= 0;
  PixelsRowIndex:= 2;
  RGBIndex:= 1;
  Inc(PixelsRow);

  {-- GET THE LENGTH OF HIDDEN DATA --}

  for BitIndex:= 0 to SizeOf(DataSize) * 8 - 1 do
    begin
    SetBitAt(DataSize, BitIndex, GetBitAt(PixelsRow^[RGBIndex], PixelBitIndex));
    CheckNextPixel;
    end;

  if DataSize <= 0 then
    raise Exception.Create('The stored size of the hidden data is not correct.' + #13#10 +
                           'This is probably caused by an image that doesn''t contain any hidden data.');

  ProgressBar.Max:= DataSize div 10;

  {-- EXTRACT THE ACTUAL DATA --}

  for DataIndex:= 1 to DataSize do
    begin

    if DataIndex mod 10 = 0 then
      begin
      ProgressBar.StepIt;
      Application.ProcessMessages;
      end;

    for BitIndex:= 0 to 7 do
      begin
      SetBitAt(Data, BitIndex, GetBitAt(PixelsRow^[RGBIndex], PixelBitIndex));
      CheckNextPixel;
      end;

    Destination.Write(Data, 1);

    end; //for DataIndex
end;

procedure Decrypt(const SourceFile, DestFile: string; ProgressBar: TProgressBar);
var Bitmap: TBitmap;
    Destination: TFileStream;

begin
  ProgressBar.Position:= 0;
  ProgressBar.Step:= 1;

  Bitmap:= TBitmap.Create;
  Destination:= nil;
  try

    try
      if FileExists(DestFile) then DeleteFile(PAnsiChar(DestFile));

      Destination:= TFileStream.Create(DestFile, fmCreate);
      Bitmap.LoadFromFile(SourceFile);

      if Bitmap.PixelFormat <> pf24bit then
        raise Exception.Create('The image doesn''t have a 24-bit depth. It surely hasn''t been created by this program.');

      ProcessDecrypt(Bitmap, Destination, ProgressBar);

    finally
      Bitmap.Free;
      if Assigned(Destination) then Destination.Free;
    end;

  except
    if FileExists(DestFile) then
      DeleteFile(PChar(DestFile));
    raise;
  end;
end;

end.
